# -*- coding: utf-8 -*-

"""Top-level package for Stochastic Simulation Algorithms in Python."""

__author__ = """Dileep Kishore, Srikiran Chandrasekaran"""
__email__ = "k.dileep1994@gmail.com"
__version__ = "1.0.1"

from .simulation import Simulation
